@@ -1,15 +1,72 @@
 from __future__ import division
 
+import healpy as hp
 import numpy as np
 from itertools import izip
 from pyoperators import (
     BlockColumnOperator, BlockDiagonalOperator, DiagonalOperator,
     MPIDistributionIdentityOperator, PackOperator, pcg, proxy_group,
     rule_manager)
-from .acquisition import QubicAcquisition
 from .utils import progress_bar
 
-__all__ = ['map2tod', 'tod2map_all', 'tod2map_each']
+__all__ = ['angular_distance_from_mask',
+           'apodize_mask',
+           'map2tod',
+           'tod2map_all',
+           'tod2map_each']
+
+
+def angular_distance_from_mask(mask):
+    """
+    For each pixel of a Healpix map, return the smallest angular distance
+    to a set of masked pixels (of value True), in degrees.
+
+    Parameter
+    ---------
+    maskok : boolean Healpix map
+        The Healpix mask that defines the set of masked pixels (of value True)
+        whose smallest angular distance to each pixel of a Healpix map of same
+        nside is computed.
+
+    """
+    nside = hp.npix2nside(len(mask))
+
+    # get the list of pixels on the external border of the mask
+    ip = np.arange(12*nside**2)[~mask]
+    neigh = hp.get_all_neighbours(nside, ip)
+    nn = np.unique(neigh.ravel())
+    if nn[0] == -1:
+        nn = nn[1:]
+    nn = nn[mask[nn]]
+
+    # get unit vectors for border and inner pixels
+    vecpix_inner = np.array(hp.pix2vec(nside, ip))
+    vecpix_outer = np.array(hp.pix2vec(nside, nn))
+
+    # get angles between the border pixels and inner pixels
+    cosang = np.dot(vecpix_inner.T, vecpix_outer)
+    mapang = np.zeros(12*nside**2)
+    mapang[~mask] = np.degrees(np.arccos(np.max(cosang, axis=1)))
+    return mapang
+
+
+def apodize_mask(maskok, fwhm_deg):
+    """
+    Apodize a mask.
+
+    Parameters
+    ----------
+    maskok : boolean Healpix map
+        The mask to be apodized (of value True).
+
+    fwhm_deg : float
+        The FWHM of the apodization function, counted from the mask edges,
+        in degrees.
+
+    """
+    sigma_deg = fwhm_deg / np.sqrt(8 * np.log(2))
+    mapang = angular_distance_from_mask(~maskok)
+    return 1 - np.exp(-0.5 * mapang**2 / sigma_deg**2)
 
 
 def _get_projection_restricted(acq, P, mask):
@@ -91,7 +148,7 @@ def map2tod(acq, map, convolution=False, max_nbytes=None):
 
 
 def _tod2map(acq, tod, coverage_threshold, disp_pmatrix, max_nbytes, callback,
-             disp_pcg, maxiter, tol):
+             disp_pcg, maxiter, tol, _factor):
     projection = acq.get_projection_operator(verbose=disp_pmatrix)
     distribution = acq.get_distribution_operator()
     P = projection * distribution
@@ -108,7 +165,17 @@ def _tod2map(acq, tod, coverage_threshold, disp_pmatrix, max_nbytes, callback,
     coverage = P.T(ones)
     if kind == 'IQU':
         coverage = coverage[..., 0]
-    mask = coverage > coverage_threshold
+    cov = coverage[coverage > 0]
+    i = np.argsort(cov)
+    cdf = np.cumsum(cov[i])
+    j = np.argmax(cdf >= coverage_threshold * cdf[-1])
+    threshold = cov[i[j]]
+    mask = coverage >= threshold
+    if acq.comm.rank == 0 and coverage_threshold > 0:
+        print 'Total coverage:', cdf[-1]
+        print 'Threshold coverage set to:', threshold
+        print 'Fraction of rejected pixels:', 1 - np.sum(mask) / cov.size
+
     projection = _get_projection_restricted(acq, projection, mask)
     pack = PackOperator(mask, broadcast='rightward')
     hwp = acq.get_hwp_operator()
@@ -118,17 +185,23 @@ def _tod2map(acq, tod, coverage_threshold, disp_pmatrix, max_nbytes, callback,
     if acq.scene.kind == 'QU':
         H = acq.get_subtract_grid_operator() * H
 
+    if _factor is not None:
+        coef = np.array([1, _factor, _factor])
+        H = H * DiagonalOperator(1 / coef, broadcast='leftward')
+
     invNtt = acq.get_invntt_operator()
     preconditioner = DiagonalOperator(1/coverage[mask], broadcast='rightward')
     A = H.T * invNtt * H
     solution = pcg(A, H.T(invNtt(tod)), M=preconditioner, callback=callback,
                    disp=disp_pcg, maxiter=maxiter, tol=tol)
+    if _factor is not None:
+        solution['x'] *= coef
     output_map = pack.T(solution['x'])
     return output_map, coverage
 
 
-def tod2map_all(acquisition, tod, coverage_threshold=0, max_nbytes=None,
-                callback=None, disp=True, maxiter=300, tol=1e-4):
+def tod2map_all(acquisition, tod, coverage_threshold=0.01, max_nbytes=None,
+                callback=None, disp=True, maxiter=300, tol=1e-4, _factor=None):
     """
     Compute map using all detectors.
 
@@ -142,8 +215,9 @@ def tod2map_all(acquisition, tod, coverage_threshold=0, max_nbytes=None,
     tod : array-like
         The Time-Ordered-Data of shape (ndetectors, ntimes).
     coverage_threshold : float, optional
-        The coverage threshold used to reject map pixels from
-        the reconstruction.
+        The low-coverage sky pixels whose cumulative coverage is below a
+        fraction of the total coverage are rejected. This keyword speficies
+        this fraction (between 0 and 1]).
     max_nbytes : int
         Maximum number of bytes for the pointing matrix. If the actual size
         is greater than this number, the computation of the pointing matrix
@@ -166,11 +240,11 @@ def tod2map_all(acquisition, tod, coverage_threshold=0, max_nbytes=None,
         with npix = 12 * nside**2
     """
     return _tod2map(acquisition, tod, coverage_threshold, True, max_nbytes,
-                    callback, disp, maxiter, tol)
+                    callback, disp, maxiter, tol, _factor)
 
 
 def tod2map_each(acquisition, tod, coverage_threshold=0, max_nbytes=None,
-                 callback=None, disp=True, maxiter=300, tol=1e-4):
+                 callback=None, disp=True, maxiter=300, tol=1e-4, _factor=None):
     """
     Compute average map from each detector.
 
@@ -214,14 +288,14 @@ def tod2map_each(acquisition, tod, coverage_threshold=0, max_nbytes=None,
     if disp:
         bar = progress_bar(len(instrument), 'TOD2MAP_EACH')
     for i, t in izip(instrument, tod):
-        acq = QubicAcquisition(i, acquisition.sampling, acquisition.scene)
+        acq = type(acquisition)(i, acquisition.sampling, acquisition.scene)
         x_, n_ = _tod2map(acq, t[None, :], coverage_threshold, False,
-                          max_nbytes, callback, False, maxiter, tol)
+                          max_nbytes, callback, False, maxiter, tol, _factor)
         x += x_
         n += n_
         if disp:
             bar.update()
 
-    if acq.scene.kind == 'I':
+    if acquisition.scene.kind == 'I':
         return np.nan_to_num(x / n), n
     return np.nan_to_num(x / n[:, None]), n
